#ifndef float_image_H
#define float_image_H

/* Multichannel images with floating-point samples. */
/* Last edited on 2007-07-01 11:15:33 by stolfi */ 

#include <stdio.h>
#include <bool.h>
#include <indexing.h>

typedef struct float_image_t 
  { ix_size_t sz[3]; /* Image sizes: sz[0] = channels, sz[1] = columns, sz[2] = rows. */
    ix_step_t st[3]; /* Indexing step along each axis. */
    ix_pos_t bp;     /* Position of sample {[0,0,0]}. */
    float *sample;   /* Linearized sample vector. */
  } float_image_t; 
  /* The sample value for channel {c}, column {x}, and row {y} 
    of the image is stored in {sample[bp + c*st[0] + x*st[1] + y*st[2]]}. */

#define float_image_max_size 65356
  /* Maximum dimension along any axis (2^16), for sanity checks. */

/* ACESSING INDIVIDUAL SAMPLES */

float float_image_get_sample(float_image_t *fim, int c, int x, int y);
  /* Returns the sample in channel {c}, column {x}, row {y} of image {fim}. */

float *float_image_get_sample_address(float_image_t *fim, int c, int x, int y);
  /* Returns the address of the sample in channel {c}, column {x}, row {y} 
    of image {fim}. */

void float_image_set_sample(float_image_t *fim, int c, int x, int y, float v);
  /* Stores {v} into the sample in channel {c}, column {x}, row {y} of image {fim}. */

/* ACESSING WHOLE PIXELS */

void float_image_get_pixel(float_image_t *fim, int x, int y, float v[]);
  /* Copies into {v[c]} the sample in channel {c}, column {x} and row
    {y} of image {fim}, for {c=0..NC-1}; where {NC} is the number of
    channels of {fim}. */

void float_image_set_pixel(float_image_t *fim, int x, int y, float v[]);
  /* Stores the sample value {v[c]}, into channel {c}, column {x} and
    row {y} of image {fim}, for {c=0..NC-1}; where {NC} is the number
    of channels of {fim}. */
    
void float_image_fill_pixel(float_image_t *fim, int x, int y, float v);
  /* Stores the sample value {v} into channel {c}, column {x} and row
    {y} of image {fim}, for {c=0..NC-1}; where {NC} is the number of
    channels of {fim}. */

/* IMAGE INTERPOLATION

  For the procedures in this section, the pixel on column {ix} and 
  row {iy} is assumed to be centered at point {(ix+0.5,iy+0.5)}. */

float float_image_interpolate_samples(float_image_t *fim, int c, double x, double y);
  /* Returns the value of channel {c} of image {fim} at the point
    {(x,y)}, computed by bilinear interpolation from nearby
    samples. 
    
    More precisely, if {x} and {y} are of the form {ix + fx} and {iy +
    fy}, with integer {ix,iy} and {fx,fy} in [-0.5 _ +0.5], then
    returns the weighted average of the four pixels in columns {ix-1}
    and {ix}, with weights {0.5-fx} and {0.5+fx}, and rows {iy-1} and
    {iy}, with weights {0.5-fy} and {0.5+fy}.
    
    For this formula, pixels outside the array bounds are assumed to
    be filled by duplicating the extremal rows and columns of
    {fim}. */

void float_image_interpolate_pixels(float_image_t *fim, double x, double y, float v[]);
  /* For {c=0..NC-1}, stores into {v[c]} the value of chanel {c} of
    image {fim} at the point {(x,y)}, computed by bilinear
    interpolation from nearby samples; where {NC} is the number of
    channels of {fim}. */

/* ACESSING WHOLE ROWS OF SAMPLES */

void float_image_get_sample_row(float_image_t *fim, int c, int y, float v[]);
  /* Copies into {v[x]} the sample of channel {c}, column {x}, row {y}
    of image {fim}, for {x=0..NX-1}; where {NX} is the number of
    columns of {fim}. */

void float_image_set_sample_row(float_image_t *fim, int c, int y, float v[]);
  /* Stores the sample value {v[x]} into channel {c}, column {x}, row {y} of
    image {fim}, for {x=0..NX-1}; where {NX} is the number of columns
    of {fim}. */
    
void float_image_fill_sample_row(float_image_t *fim, int c, int y, float v);
  /* Stores the sample value {v} into channel {c}, column {x}, row {y}
    of image {fim}, for {x=0..NX-1}; where {NX} is the number of
    columns of {fim}. */

/* ACESSING WHOLE ROWS OF PIXELS */

void float_image_get_pixel_row(float_image_t *fim, int y, float v[]);
  /* Copies into {v[c + x*NC]} the sample of channel {c}, column {x},
    row {y} of image {fim}, for {c=0..NC-1} and {x=0..NX-1}; where
    {NC} and {NX} are the number of channels and the number of columns
    of {fim}. */

void float_image_set_pixel_row(float_image_t *fim, int y, float v[]);
  /* Stores the sample value {v[c + x*NC]} into channel {c}, column
    {x}, row {y} of image {fim}, for {c=0..NC-1} and {x=0..NX-1};
    where {NC} and {NX} are the number of channels and the number of
    columns of {fim}. */
    
void float_image_fill_pixel_row(float_image_t *fim, int y, float v);
  /* Stores the sample value {v} into channel {c}, column {x}, row {y}
    of image {fim}, for {c=0..NC-1} and {x=0..NX-1}; where {NC} and {NX}
    are the number of channels and the number of columns of {fim}. */

/* ACCESSING WHOLE CHANNELS */

void float_image_fill_channel(float_image_t *fim, int c, float v);
  /* Stores {v} into all samples in channel {c} of image {fim}. */

/* ACCESSING THE WHOLE IMAGE */

void float_image_fill(float_image_t *fim, float v);
  /* Stores {v} into all samples of image {fim}. */

/* ALLOCATION */

float_image_t *float_image_new (int NC, int NX, int NY);
  /* Allocates a new multichannel image with {NC} channels,
    {NX} columns, and {NY} rows. */

void float_image_free(float_image_t *fim);
  /* Discards the pixel array and header of the image {fim}. */

/* SAMPLE RANGE */

void float_image_update_sample_range(float_image_t *fim, int c, float *vmin, float *vmax);
  /* Expands the range {[*vmin _ *vmax]} so that it contains all
    sample values in channel {c} of image {fim}.  If {c} is not a valid
    channel index, assumes that all samples in that channel are all zero. */

/* GET/CHECK SIZE */
  
void float_image_get_size(float_image_t *IMG, int *NC, int *NX, int *NY);
  /* Stores the number of channels, columns, and rows of {IMG} in
    {*NC}, {*NX}, and {*NY}, respectively. */
  
void float_image_check_size(float_image_t *IMG, int NC, int NX, int NY);
  /* Bombs out if {IMG} does not have exactly {NC} channels, {NX}
    columns, and {NY} rows. */

/* INPUT/OUTPUT */

void float_image_write(FILE *wr, float_image_t *fim);
  /* Writes the image {fim} to {wr}, in ASCII floating point format.
    The file will start with a standard header line 
    "begin float_image_t (version {DATE})", 
    followed by three lines with "NC = {NUM}", "NX = {NUM}", "NY =
    {NUM}". Then follow the pixels, one per line with "{X} {Y}
    {SAMPLE}...". A blank line is inserted between two consecutive
    rows. The file ends with a footer "end float_image_t". */

float_image_t *float_image_read(FILE *rd);
  /* Reads an image from {rd}, in ASCII floating point format
    generated by {float_image_write}. */

#endif
