/* See {sve_minn.h} */
/* Last edited on 2007-10-28 18:24:56 by stolfi */

#include <sve_minn.h>

#include <gauss_elim.h>
#include <bool.h>
#include <vec.h>
#include <rmxn.h>
#include <rmxn_extra.h>

#include <math.h>

void sve_minn_step(int n, double fv[], double cm[])
  { int rows = n+2; /* {n+1} stationary eqs, and one unit-sum eq. */
    int cols = n+3; /* {n+1} barycentric coords, one Lagrange multip, and the indep term. */
    double M[rows*cols];
    int i, j;
    /* Fill in the main submatrix: */
    int jb = cols - 1; /* Column of independent term. */
    for (i = 0; i <= n; i++)
      { for (j = 0; j <= n; j++)
          { double fij = (j <= i ? fv[i*(i+1)/2 + j] : fv[j*(j+1)/2 + i]);
            M[i*cols + j] = fij;
            if (i == j) { M[i*cols + jb] = fij/4; }
          }
      }
    /* Fill in the last row and column with the unit-sum constraint: */
    int ije = n+1; /* Index of constraint row & column. */
    for (i = 0; i <= n; i++) { M[i*cols + ije] = 1; M[ije*cols + i] = 1; }
    M[ije*cols + ije] = 0;
    M[ije*cols + jb] = 1;
    /* Solve the system: */
    gsel_triangularize(rows, cols, M, TRUE);
    gsel_diagonalize(rows, cols, M, TRUE);
    gsel_normalize(rows, cols, M, TRUE);
    double x[n+2];
    int r = gsel_extract_solution(rows, cols, M, 1, x, TRUE);
    if (r < rows) 
      { fprintf(stderr, "%s: warning - indeterminate system, r = %d\n", __FUNCTION__, r); }
    /* Extract the solution: */
    double sum = 0.0;
    for (i = 0; i <= n; i++) { cm[i] = x[i]; sum += cm[i]; }
    if (fabs(sum - 1.0) > 0.5e-7) 
      { fprintf(stderr, "%s: warning - normalization failed, sum = %24.16e\n", __FUNCTION__, sum); }
    /* Just to be sure: */
    if ((sum != 0) && (sum != 1)) { for (i = 0; i <= n; i++) { cm[i] /= sum; } }
  }

void sve_sample_function(int n, int m, sve_goal_t *F, double v[], double fv[])
  { double x[m];
    int i, j;
    for (i = 0; i <= n; i++)
      { for (j = 0; j <= i; j++)
          { /* Set {x[0..n-1]} to the midpoint of simplex corners {i,j}: */
            int k; for (k = 0; k < m; k++) { x[k] = (v[i*m + k] + v[j*m + k])/2; }
            /* Get the function's value {F(x)} at {x}, store {F(x)} into {fv}: */
            fv[i*(i+1)/2 + j] = F(m, x);
          }
      }
  }

void sve_minn_iterate
  ( int n, 
    sve_goal_t *F, 
    sve_pred_t *OK,
    double x[], 
    double rIni,
    double rMin, 
    double rMax,
    double dMax,
    int maxEvals
  )
  { /* Save first guess: */
    double x0[n];   /* Cartesian coords of initial guess. */
    int i; 
    for (i = 0; i < n; i++) { x0[i] = x[i]; }
    
    /* Allocate the working storage for optimization: */
    int nv = n+1;
    double cm[nv]; /* Barycentric coords of next solution. */
    double_vec_t vv = double_vec_new(nv*n);
    double *v = vv.el; /* Probe simplex: */
    int nf = (n+1)*(n+2)/2;
    double_vec_t fvv = double_vec_new(nf);
    double *fv = fvv.el; /* Sampled function values. */

    /* Iteration: */
    double y[n];    /* Cartesian coords of next solution. */
    double radius = rIni; /* Current probe simplex radius: */
    double dPrev = -1; /* Distance moved in previous iteration (-1 if none). */
    int iter = 0;   /* Counts main iterations. */
    int nEvals = 0; /* Counts calls to {F}. */
    bool_t debug = FALSE;  /* Set it to TRUE to print debugging info. */
    while(TRUE) 
      { fprintf(stderr, "iteration %4d radius = %12.8f\n", iter, radius);
        /* Evaluate function at current solution, just to check progress: */
        if (debug) 
          { rn_gen_print(stderr, n, x, "%12.8f", "  center = \n    [ ", "\n      ", "   ]\n"); }
        double Fx = F(n, x);
        if (debug)
          { fprintf(stderr, "  function = %22.16e\n", Fx); }
        /* Check for termination: */
        if ((OK != NULL) && (OK(n, x, Fx))) { break; }
        /* Choose a simplex around {x}: */
        rmxn_regular_simplex(n, v);
        rmxn_spin_rows(n+1, n, v, v);
        double sf = radius/rmxn_regular_simplex_radius(n);
        rmxn_scale(n+1, n, sf, v, v);
        rmxn_shift_rows(n+1, n, v, x, v);
        /* Evaluate the goal function at sample points: */
        sve_sample_function(n, n, F, v, fv);
        /* Optimize as a quadratic: */
        sve_minn_step(n, fv, cm);
        /* Convert solution to Cartesian coordinates: */
        rmxn_map_row(n+1, n, cm, v, y);
        /* Enforce the limit {dMax} for the total motion: */
        double d0 = rn_dist(n, x0, y);
        if (d0 > dMax) 
          { /* Moved too much, curb it: */
            double s = dMax/d0;
            for (i = 0; i < n; i++) { y[i] = x0[i] + s*(y[i] - x0[i]); }
          }
        /* Check motion of this iteration: */
        double d = rn_dist(n, x, y);
        double stepMax = 8*radius; /* Max move in this iteration. */
        if (d > stepMax)
          { /* Moved too much, curb it: */
            double s = stepMax/d;
            for (i = 0; i < n; i++) { y[i] = x[i] + s*(y[i] - x[i]); }
            d = stepMax;
          }
        /* Update solution: */
        for (i = 0; i < n; i++) { x[i] = y[i]; }
        /* Check budget: */
        nEvals += nf + 1;
        if (nEvals > maxEvals) { return; }
        /* Check for convergence: */
        if ((dPrev > d) && (d*d < rMin*(dPrev - d))) { return; }
        /* Remember how much we moved in this iteration: */
        dPrev = d;
        /* Adjust radius for next iteration: */
        double rNew = d/2;
        if (rNew < 0.5*radius) { rNew = 0.5*radius; }
        if (rNew > 2.0*radius) { rNew = 2.0*radius; }
        radius = rNew;
        if (radius < rMin) { radius = rMin; }
        if (radius > rMax) { radius = rMax; }
        iter++;
      }
    free(v); free(fv);
  }

