/* test_sve_catenary --- test of {sve_minn.h} for hanging-chain energy minimization. */
/* Last edited on 2007-10-30 07:31:27 by stolfi */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <math.h>
#include <values.h>

#include <sve_minn.h>

#include <bool.h>
#include <argparser.h>
#include <jsrandom.h>
#include <jsstring.h>
#include <affirm.h>
#include <rmxn.h>
#include <rmxn_extra.h>
#include <rn.h>
#include <jsfile.h>
#include <vec.h>

#include <fftw3.h>

/* GENERAL PARAMETERS */

typedef struct options_t
  { int nk; /* Number of links in the chain. */
    double wd; /* Ideal distance between attachment points. */
  } options_t;

#define MAXLINKS 256
  /* Max number of links in chain. */

/* INTERNAL PROTOTYPES */

int main (int argc, char **argv);

options_t *get_options(int argc, char **argv);
  /* Parses the command-line options. */

void find_chain_shape(int nk, double wd);
  /* Tries to find node positions that minimize the 
    potential energy of an {nk}-link chain with ideal distance
    between endpoints close to {wd}. */ 

void write_node_positions(char *fname, int nk, double x[], double y[]);
  /* Writes the chain node positions to a file named "{fname}". Each
    line contains the node index and two values {(x[i],y[i])} for {i =
    0..nk}.  Note that {x} and {y} must have {nk+1} elements each.
    If {fname} is NULL, writes to {stderr} instead. */

void compute_node_positions(int nk, double c[], double x[], double y[]);
  /* Computes the node positions implied by the coefficients
    {c[0..nk-2]}. Coefficient {c[i]} is the angle between links
    {i} and {i+1} of the chain (angle zero means the links
    are aligned, positive means an upward bend).
    
    The arrays {x,y} must have {nk+1} elements each. Each pair
    {(x[j],y[j])} will be the coordinates of articulation {j} from the
    left. In particular, the point {(x[0],y[0])} is always {(0,0)} and
    point {(x[nk-1],y[nk-1])} is {(a,0)} for some {a}. */

double chain_energy(int nk, double x[], double y[], double wd);
  /* The potential energy of the chain, plus the elastic energy of a spring
    that tries to keep the right end-point of the chain at abscissa {x[nk] = 1}. */

/* IMPLEMENTATIONS */

int main (int argc, char **argv)
  { options_t *o = get_options(argc, argv);
    srand(4615);  srandom(4615);
    find_chain_shape(o->nk, o->wd);
    fclose(stderr);
    fclose(stdout);
    return (0);
  }
  
void find_chain_shape(int nk, double wd)
  { 
    /* Number of minimization variables: */
    int nc = nk - 1;
    
    /* Output file names: */
    char *prefix = NULL; /* Prefix for output file names. */
    asprintf(&prefix, "test-%03d", nk);
    char *inifname = txtcat(prefix, "-ini-ot.dat"); /* Name of initial guess file. */
    unlink(inifname);
    char *solfname = txtcat(prefix, "-sol-ot.dat"); /* Name of final solution file. */
    unlink(solfname);
    
    /* Working storage for the goal function: */
    double x[nk+1], y[nk+1];  /* Node positions. */
    
    auto double F(int n, double c[]); 
      /* The goal function for optimization. */
      
    double F(int n, double c[])
      { assert(n == nc);
        /* Compute the pulse's power spectrum after padding to {2*n} samples: */
        compute_node_positions(nk, c, x, y);
        double Fc = chain_energy(nk, x, y, wd);
        return Fc;
      }
      
    double cprev[nc]; /* Guess in previous call of {OK} function. */
    int nok = 0;        /* Counts iterations (actually, calls to {OK}). */
    
    auto int OK(int n, double c[], double Fc); 
      /* Acceptance criterion function. */
      
    int OK(int n, double c[], double Fc)
      { assert(n == nc);
        fprintf(stderr, "iteration %d\n", nok);
        if (nok > 0)
          { double d = rn_dist(n, cprev, c);
            fprintf(stderr, "change = %16.10f\n", d);
          }
        fprintf(stderr, "chain =\n");
        compute_node_positions(nk, c, x, y);
        write_node_positions(NULL, nk, x, y);
        /* Save guess in {xprev} for next call: */
        rn_copy(n, c, cprev); 
        nok++;
        fprintf(stderr, "potential = %16.10f\n", Fc);
        fprintf(stderr, "\n");
        return 0;
      }

    double c[nc];     /* Initial guess and final solution. */
    
    /* Initialize {c} with a semicircle: */
    int i;
    double c0 = M_PI/nc; 
    for (i = 0; i < nc; i++) { c[i] = c0; }
    
    /* Print and write the initial guess: */
    fprintf(stderr, "initial guess:\n");
    compute_node_positions(nk, c, x, y);
    write_node_positions(NULL, nk, x, y);
    write_node_positions(inifname, nk, x, y);
    /* Print the goal function's value: */
    fprintf(stderr, "initial potential = %+24.16e\n", F(nc, c));

    /* Optimize iteratively: */
    int NP = (nc+1)*(nc+2)/2; /* Number of probe points per iteration. */
    double rMin = 0.0000001;
    double rMax = 0.10*M_PI;
    double rIni = 0.02*M_PI;
    int NITER = 200;
    int maxEvals = (NP+1)*NITER;
    sve_minn_iterate(nc, &F, &OK, c, rIni, rMin, rMax, MAXDOUBLE, maxEvals);
    
    /* Print and write final solution: */
    fprintf(stderr, "final solution:\n");
    compute_node_positions(nk, c, x, y);
    write_node_positions(NULL, nk, x, y);
    write_node_positions(solfname, nk, x, y);
    /* Print the goal function's value: */
    fprintf(stderr, "final potential = %+24.16e\n", F(nc, c));
  }

void write_node_positions(char *fname, int nk, double x[], double y[])
  { FILE *wr = (fname != NULL ? open_write(fname, TRUE) : stderr);
    int i;
    for (i = 0; i <= nk ; i++)
      { fprintf(wr, "%5d %12.8f %12.8f\n", i, x[i], y[i]); }
    if ((wr != stderr) && (wr != stdout)) { fclose(wr); }
  }

void compute_node_positions(int nk, double c[], double x[], double y[])
  { int nc = nk - 1;
    /* Initial node is always {(0,0)}: */
    x[0] = y[0] = 0;
    /* Assume that link 0 is horizontal: */
    double a = 0;        /* Angle of next link. */
    x[1] = 1; y[1] = 0; 
    /* Compute {(x[i],y[i]) for {i} in {2..nk}: */
    int i;
    for (i = 0; i < nc; i++)
      { a = a + c[i];
        double dx = cos(a), dy = sin(a);
        x[i+2] = x[i+1] + dx;
        y[i+2] = y[i+1] + dy;
      }
    /* Compute the rotation needed to bring the last node to the {x}-axis: */
    double r = - atan2(y[nk], x[nk]);
    double cr = cos(r), sr = sin(r);
    for (i = 1; i <= nk; i++)
      { double xr = cr*x[i] - sr*y[i];
        double yr = sr*x[i] + cr*y[i];
        x[i] = xr; y[i] = yr;
      }
  }

double chain_energy(int nk, double x[], double y[], double wd)
  { /* Compute the gravitational energy {G}: */
    double G = 0;
    int i;
    for (i = 0; i < nk; i++) 
      { double yb = (y[i] + y[i+1])/2;
        G += yb;
      }
    /* Compute elastic energy {E} of spring at right end: */
    double dx = x[nk] - wd;
    double E = dx*dx;
    /* Compute the total energy: */
    return G + E;
  }
  
options_t *get_options(int argc, char **argv)
  { argparser_t *pp = argparser_new(stderr, argc, argv);
    options_t *o = notnull(malloc(sizeof(options_t)), "no mem"); 
    o->nk = argparser_get_next_int(pp, 1, MAXLINKS);
    o->wd = argparser_get_next_double(pp, 0.0, MAXDOUBLE);
    argparser_finish(pp);
    return o;
  }
