// See minu_brent.h
// Last edited on 2007-01-04 03:44:17 by stolfi

#include <minu_gen.h>
#include <minu_brent.h>

#include <affirm.h>

#include <values.h>
#include <stdio.h>

#define ABS(x) ((x) < 0 ? -(x) : (x))
#define MAX(x,y) ((x) > (y) ? (x) : (y))
#define MIN(x,y) ((x) < (y) ? (x) : (y))

#define Eps (1.0e-6) /* Square root of relative machine precision. */
#define Phi    (1.61803398874989484821) /* Golden ratio */
#define Ihp    (0.61803398874989484821) /* 1/Phi */
#define IhpSqr (0.3819660112501051518)  /* 1/Phi^2 */
 
// INTERNAL PROTOTYPES

void mub_message (char *msg);
void mub_newline(void);
void mub_print_abscissa(char *msg, double u);
void mub_print_point(char *msg, double u, double fu);

// IMPLEMENTATIONS

void minu_brent_minimize
  (
    void *parms,
    EvalProc eval,
    double *x,
    double *fx,
    double *dfx,
    double tol,
    double dist,
    double *a,
    double *b,
    CheckProc check,
    bool_t debug
  )
{
  double v, fv;        /* Third-smallest point of current triple */
  double w, fw;        /* Second-smallest point of current triple */
  double u, fu;        /* Splitting point */

  #define PrintStatus() \
    {  \
      mub_newline(); \
      mub_print_abscissa("a", *a); \
      mub_print_point("v", v, fv); \
      mub_print_point("x", *x, *fx); \
      mub_print_point("w", w, fw); \
      mub_print_abscissa("b", *b); \
      mub_newline(); \
    }

  double d, e, r;      /* Step sizes */
  double Dfx,Dq;       /* Estimated derivative at `x' is `Dfx/Dq'. */
  double DDfx,DDq;     /* Estimated second derivative at `x' is `DDfx/DDq'. */
  double dfu;          /* Derivative at "u" (returned but ignored). */
  double fxOld;        /* Previous value of "fx" */
  
  double tol1, tol2;
  double tol3 = tol/3.0; 
  bool_t aInf, bInf;

  affirm(*a <= *x, "");
  affirm(*x <= *b, "");
  if (check != NULL)
    { if (check(parms, *a, *b, *x, *fx, 0.0,0.0, 0.0,0.0)){ return; } }

  w = *x; fw = *fx;
  v = *x; fv = *fx;
  e = 0.0;
  fxOld = *fx;
  while (1)
  {
    affirm(*fx <= fw, "");
    affirm(fw <= fv, "");
    aInf = (*a == -MAXDOUBLE);
    bInf = (*b == MAXDOUBLE);
    tol1 = Eps * ABS(*x) + tol3;
    tol2 = 2.0 * tol1;
    if (aInf){ affirm(((*x) <= w) && (w <= v), ""); }
    if (bInf){ affirm(((*x) >= w) && (w >= v), ""); }

    if (debug){ PrintStatus(); }

    /* check stopping criterion: */
    if ((! aInf) && (! bInf) && (MAX(x-a, b-x) <= tol2)){ return; }

    if (ABS(e) > tol1)
      { /* Fit parabola through "v", "x", "w"; */
        ComputeDerivatives(v, fv, w, fw, *x, *fx, &Dfx,&Dq, &DDfx,&DDq);
        r = e; e = d;
      }
    else
      { Dfx = 0.0; Dq = 0; DDfx = 0.0; DDq = 0.0;
        r = 0.0;
      }

    /* Compute the displacement "d" from "x" to the next probe: */
    if (DDfx < 0.0){ Dfx = -Dfx; DDfx = -DDfx; }
    if (
       (ABS(Dfx)*DDq < 0.5*ABS(DDfx*r)*Dq) && 
       (aInf || (Dfx*DDq < DDfx*(x - a)*Dq)) && 
       (bInf || (Dfx*DDq > DDfx*(x - b)*Dq))
    ) 
      { /* A parabolic-interpolation step: */
        if (debug){ mub_message("parabolic step"); }
        d = -(Dfx/DDfx)*(DDq/Dq);
      }
    else if (bInf || aInf)
      { /* A golden-section extrapolation: */
        if (debug){ mub_message("golden extension"); }
        e = ((*x) != w ? (*x) - w : ((*x) != v ? (*x) - v : 0.0));
        d = (e == 0.0 ? dist/Phi : e*Phi);
        if (ABS(d) < tol1) { d = (d < 0.0 ? -tol1 : tol1); }
      }
    else
      { /* A golden-section interpolation: */
        if (debug){ mub_message("golden section"); }
        e = (*b - *x > *x - *a ? *b - *x : *a - *x);
        d = IhpSqr * e;
      }

    /* The function must not be evaluated too close to "a" or "b": */
    u = (*x) + d;
    if ((! aInf) && (u - *a < tol2))
      { d = tol1; }
    else if ((! bInf) && (*b - u < tol2))
      { d = -tol1; }

    /* The function must not be evaluated too close to x: */
    if (ABS(d) >= tol1)
      { u = (*x) + d; }
    else if (d > 0.0)
      { u = (*x) + tol1; }
    else
      { u = (*x) - tol1; }  

    if (eval(parms, u, &fu, &dfu)){ return; }

    /* update  "a", "b": */
    if (fu >= *fx)
      { if (u < (*x)) { *a = u; } else { *b = u; } }
    if (fu <= *fx)
      { if (u < (*x)) { *b = (*x); } else{ *a = (*x); } }

    /* Update "x", "v", "w": */
    affirm(*a <= u, "");
    affirm(u <= *b, "");
    if ((fu <= *fx))
      { if ((v != w) && (w != (*x))) { v = w; fv = fw; }
        w = *x; fw = *fx;
        *x = u; *fx = fu; *dfx = dfu;
        if (check != NULL)
          { if (check(parms, *a, *b, *x, *fx, Dfx,Dq, DDfx,DDq)) { return; } }
      }
    else if (fu <= fw)
      { affirm(w != *x, "");
        v = w; fv = fw;
        w = u; fw = fu;
      }
    else if (fu <= fv)
      {
        affirm(v != *x, "");
        if (w == *x)
          { w = u; fw = fu; }
        else
          { v = u; fv = fu; }
      }
    else if (v == w)
      { v = u; fv= fu; }
  }
}
void mub_message (char *msg)
{
  fprintf(stderr, "%s\n", msg);
}

void mub_newline()
{
  fprintf(stderr, "\n");
}

void mub_print_abscissa(char *msg, double u)
{ 
  fprintf(stderr, "    %s = %20g\n", msg, u);
}

void mub_print_point(char *msg, double u, double fu)
{ 
  fprintf(stderr, "    %s = %20g F(%s) = %20g\n", msg, u, msg, fu);
}
