#define PROG_NAME "test_interp"
#define PROG_DESC "test of dyadic grids: interpolation"
#define PROG_VERS "1.0"

/* Last edited on 2007-10-28 19:54:01 by stolfi */

#define test_interp_COPYRIGHT "Copyright  2007 by the State University of Campinas (UNICAMP)."

#define PROG_HELP \
  "  " PROG_NAME " \\\n" \
  "    -dDim {DDIM} \\\n" \
  "    -fDim {FDIM} \\\n" \
  "    -continuity {CONT} \\\n" \
  "    [ -superior | -inferior ] \\\n" \
  "    [ -plot ] \\\n" \
  "    -outName NAME \\\n" \
  "    " argparser_help_info_HELP " \\\n" \
  "    < {INFILE}"

#define PROG_INFO \
  "NAME\n" \
  "  " PROG_NAME " - " PROG_DESC "\n" \
  "\n" \
  "SYNOPSIS\n" \
  PROG_HELP "\n" \
  "\n" \
  "DESCRIPTION\n" \
  "  This program takes a set of points {p[0..n-1]}, corresponding" \
  " data values {f[0..n-1]}, and computes a dyadic" \
  " spline {h} such that {h(p[k]) = f[k]} for all {k}.\n" \
  "\n" \
  "  Each sample point must be a vertex of the {d}-dimensional" \
  " dyadic multigrid, specified by the index {c[k]} of its superior" \
  " cell. Note that this input also specifies a rank {r[k]} for" \
  " each point {p[k]}.  The spline {h} is defined over the smallest" \
  " finite dyadic grid {G} where every data point {p[k]} is a" \
  " complete vertex in layer {r[k]} of {G}.\n" \
  "\n" \
  "  The interpolant has the form\n" \
  "\n" \
  "    {h = SUM { u[i]*B[i] : i = 0..m-1 }}       (1)\n" \
  "\n" \
  "  where {B[0..m-1]} consists of all the maximal (or minimal)" \
  " polynomial tents of continuity order {CONT} over {G} that" \
  " are centered on the vertices {p[0..n-1]} and have" \
  " ranks {r[0..n-1]}. The coefficients {u[i]} are found by" \
  " the ordered interpolation algorithm (see C.Cardoso's thesis).\n" \
  "\n" \
  "OPTIONS\n" \
  "  -dDim {DDIM}\n" \
  "    Specifies the domain's dimension as {DDIM}.\n" \
  "\n" \
  "  -fDim {FDIM}\n" \
  "    Specifies the range's dimension as {FDIM}.\n" \
  "\n" \
  "  -continuity {CONT}\n" \
  "    Specifies the continuity {CONT} of the interpolating splines.\n" \
  "\n" \
  "  -superior\n" \
  "  -inferior\n" \
  "    These two mutually exclusive flags specify the preferred" \
  " support size of the basis elements: widest possible or" \
  " narrowest possible, respectively.  if neither is specified," \
  " assumes \"-inferior\".\n" \
  "\n" \
  "  -plot\n" \
  "    If this option is present, the program will generate a plot of" \
  " the interpolated function.\n" \
  "\n" \
  "  -outName {OUT_NAME}\n" \
!  "    Specifies the common prefix {OUT_NAME} of all output file names.\n" \
  "\n" \
  "DOCUMENTATION OPTIONS\n" \
  argparser_help_info_HELP_INFO "\n" \
  "\n" \
  "SEE ALSO\n" \
  "  test_approx(1).\n" \
  "\n" \
  "AUTHOR\n" \
  "  Started 2005-06-05 or earlier by Jorge Stolfi, IC-UNICAMP.\n" \
  "\n" \
  "MODIFICATION HISTORY\n" \
  "  2007-10-14 updated for new dyadic grid libs, by J. Stolfi.\n" \
  "\n" \
  "WARRANTY\n" \
  argparser_help_info_NO_WARRANTY "\n" \
  "\n" \
  "RIGHTS\n" \
  "  " test_approx_COPYRIGHT ".\n" \
  "\n" \
  argparser_help_info_STANDARD_RIGHTS

/* Datapoint interpolation with 2D dyadic splines. */

#include <stdio.h>
#include <math.h>
#include <dg_grid.h>
#include <dg_spline.h>
#include <dg_tree.h>
#include <dg_pulse.h>
#include <dg_locus.h>
#include <dg_tent.h>
#include <dg_tent_basis.h>
#include <bz_basic.h>
#include <dg_plot.h>
#include <dg_plot_2D_func.h>
#include <dg_makegrid.h>
#include <affirm.h>
#include <pswr.h>
#include <pswr_color.h>
#include <pswr_iso.h>
#include <test_interp.h>

typedef struct tin_options_t
  { dg_dim_t dDim;        /* Dimension of domain space. */
    dg_dim_t fDim;        /* Dimension of range space. */
    dg_cont_t c;          /* Continuity order of splines. */
    bool_t superior;      /* TRUE uses a wide basis, FALSE (default) narrow. */
    /* Output options: */
    bool_t plot;           /* TRUE to generate function and error plots. */
    char *outName;        /* Prefix for output file names. */
  } tin_options_t;

int main(int argc, char **argv);
  
tin_options_t *tin_get_options(int argn, char **argc);

void tin_read_data
  ( FILE *rd,
    int np,
    dg_dim_t dDim,
    int nf,
    dg_tree_t G, 
    dg_locus_vec_t *L, 
    double_vec_t *f
  );
  /* Reads from {rd} a list of {np} sampling points, each with {dDim}
    coordinates and {nf} data values.
    
    Stores into {L[i]} the vertex locus corresponding to the sampling
    point {i}. Also expands {G} as needed so that locus {L[i]} is a
    complete vertex.
    
    Stores into {f[i*nf + r]} the data value number {r}
    associated with the sampling point point {L[i]}. */

void tin_build_spline
  ( dg_tent_vec_t B,
    double fDim,
    double_vec_t u, 
    dg_tree_t G
  );
  /* Given a basis {B[0..n-1]} of spline elements defined over the 
    dyadic grid {G}, and a list of coefficients {coef[0..n-1]},
    computes the spline {f = SUM{ coef[i]*B[i] : i in 0..n-1 }}.
    Each patch of the spline is stored into a newly allocated 
    {bz_patch_t} record, whose address is stored into the {bz} field
    of the corresponding node of {G}. */

double tin_tol(double x, double y);
  /* Maximum box radius allowed near {p}. */ 

double tin_inside(double x, double y, double rad);
  /* TRUE if a box of radius {rad} centered at {p} intersects the 
    desired domain. */ 

double tin_dsoft(double dx, double dy, double eps);
  /* Length of {(dx,dy)}, cooked to be at least {eps}. */

void tin_test_interp(char *name, bz_patch_rdim_t fDim, dg_rank_t treeDp, dg_rank_t plotDp);
  /* Tests of plotting routines, with a root patch of degree {g},
    a random tree of maximum depth {treeDp}, and maximum plotting split 
    depth {plotDp}. */

/* Max dimension of grid's domain: */
#define MAX_DDIM (4)  

/* Max dimension of function's range space: */
#define MAX_FDIM (4)  

/* TYPES */

int main(int argc, char **argv)
  { 
    tin_options_t *o = tin_get_options(argc, argv);
    int dDim = o->dDim;
    int fDim = o->fDim;
    int c = o->continuity;
    
    /* The grid's tree: */
    dg_tree_node_t *G = tin_make_trivial_tree();
    
    /* List of loci (vertices + depths) for interpolation: */
    dg_locus_vec L;
    
    /* The function values: */
    double_vec_t f = double_vec_new(0);
    
    /* Number of function values and derivatives per range layer: */
    int nd = ipow(c + 1, dDim);
    
    /* Number of data values per sampling point: */
    int nf = nd*fDim;
    
    /* Read the sampling points {L}, function data {f}, and expand the tree {G}: */
    tin_read_data(stdin, dDim, nf, G, &L, &f);
    assert(f.nel == nf*L.nel);
      
    /* Sort loci (and data) by increasing depth: */
    tin_sort_loci(nf, &L, &f);

    /* Build a Hermite tent basis family of the right continuity and degree: */
    int g = 2*c + 1;
    dg_pulse_family_t fam[dDim];
    for (i = 0; i < dDim; i++) { fam[i] = dg_pulse_family(DG_TK_H, c, g); }
    assert(nd == dg_tent_mother_count(dDim, fam));

    /* Build a tent basis with one set of tents at each sampling vertex: */
    dg_tent_vec_t B = dg_tent_basis(dDim, fam, L);
    assert(B.nel == nd*L.nel);
    assert(f.nel == fDim*B.nel);
    
    /* Coefficient vector: */
    double u[fDim*B.nel];
    /*  {u[i*fDim + r]} is the coefficient of tent {B[i]} for component {r} of the function. */

    /* Interpolate the function values at the given points: */
    tin_interpolate(dDim, fDim, fam, f, n, &B, u);
    
    tin_build_spline(dDim, fDim, fam, &B, u, G);
    
    /* Plot the result: */
    PSStream *ps = dg_plot_init(name, epsformat, "a4", "Plot test", bbox);
    dg_plot_2D_func(ps, &b, 0, NULL, G, o->plotDp, o->isolines); }
    dg_plot_finish(ps);
    
    return 0;
  }
  
void tin_read_data
  ( FILE *rd,
    int np,
    dg_dim_t dDim,
    int nf,
    dg_tree_t G, 
    dg_locus_vec_t *L, 
    double_vec_t *f
  )   
  {
    int i;
    (*L) = dg_locus_vec_new(np);
    (*f) = double_vec_new(np*nf);
    
    for (i - 0; i < np; i++)
      { /* Read a sampling point and its function values: */
        tin_read_sample(rd, dDim, &(L.el[i]), nf, &(f[i*nf]);
        /* Make sure it is a complete vertex in {G}: */
        tin_make_locus_complete(G, L.el[i]);
      }
  }
  
void tin_interpolate
  ( dg_dim_t dDim, 
    dg_pulse_family_t fam[],  
    double f[],  
    int n,  
    dg_tent_vec_t *B,  
    double u[] 
  )
  {
    
    /* Interpolation algorithm: */
    for (i = 0; i < L.nel; i++)
      { 
        /* Get the locus (vertex) {Li}: */
        dg_locus_t Li = L.el[i];
        
        /* Get the root-relative bbox {B} of its superior cell: */
        interval_t B[dDim];
        dg_cell_box_root_relative(dDim, Li.cell, B);
        
        /* Extract the lower corner {x} (root-relative vertex coords): */
        double x[dDim];
        for (r=0; r < dDim; r++) { x[i] = LO(B[i]); }
        
        /* Get the function values {fi} specified for it: */
        double *fi = &(f[i*nf]);
        
        /* Add values and derivatives {s[...]} of all tents for all prev loci: */
        double s[nf];
        for (kr = 0; kr < nf; kr++) { s[kr] = 0; }
        for (j = 0; j < i; j++)
          { 
            for (tix = 0; tix < nd; tix++)
              { 
                dg_tent_t t = B.el[j*nd + tix];
                assert(t.tix = tix);
                assert(t.cell = Li.cell);
                /* Coefficients for this tent: */
                double *ujt = &(u[j*nf + tix*fDim]);
                /* Get value and all derivs of {t} at {x}: */
                double v[tDim];
                dg_tent_eval_diff_root_relative(dDim, DG_TK_H, c, g, t, x, v);
                /* Accumulate on {s}: */
                for (k = 0; k < tDim, k++)
                  { 
                    for (r = 0; r < fDim; r++)
                      { 
                        s[k*fDim + r] += ujt[r] * v[k*fDim + r];
                      }
                  }
              }
        
          }
          
        /* Assume that the tents are cardinal with respect to derivative: */
        /* Subtract {s} from specified values {fi}, result is coeffs {u}: */
        for (kr = 0; kr < nf; kr++) { ui[kr] = fi[kr] - s[kr]; }
      }
    
  }
  
tin_options_t *tin_get_options(int argn, char **argc)
  {
    argparser_t *pp = argparser_new(stderr, argc, argv);
    
    argparser_set_help(pp, PROG_NAME " version " PROG_VERS ", usage:\n" PROG_HELP);
    argparser_set_info(pp, PROG_INFO);
    argparser_process_help_info_options(pp);
     
    tin_options_t *o = (tin_options_t *)notnull(malloc(sizeof(tin_options_t)), "no mem");

    argparser_get_keyword(pp, "-dDim");                               
    o->dDim = argparser_get_next_int(pp, 1, dg_MAX_DIM);  
                                                 
    argparser_get_keyword(pp, "-fDim");                               
    o->fDim = argparser_get_next_int(pp, 1, bz_MAX_DIM);  
                                                 
    argparser_get_keyword(pp, "-continuity");                               
    o->degree = argparser_get_next_int(pp, 1, dg_MAX_DEGREE);  
    
    if (argparser_keyword_present(pp, "-superior"))
      { o->superior = TRUE; }
    else if (argparser_keyword_present(pp, "-inferior"))
      { o->superior = FALSE; }
    else 
      { o->superior = TRUE; }
      
    o->plot = argparser_keyword_present(pp, "-plot");
                                                 
    argparser_get_keyword(pp, "-outName");                               
    o->outName = argparser_get_next(pp);  

    /* Check for spurious args: */
    argparser_finish(pp);

    return o;
  }

